package org.jvnet.annox.model;

import java.lang.annotation.Annotation;
import java.lang.reflect.Array;

import org.apache.commons.lang.Validate;
import org.jvnet.annox.util.ArrayUtils;

/**
 * Defines an annotation field.
 * 
 * @param <T>
 *            Type of the field.
 * @author Aleksei Valikov
 */
public abstract class XAnnotationField<T> {

	/**
	 * Name of the field.
	 */
	private final String name;

	/**
	 * Returns the name of the field.
	 * 
	 * @return Name of the field.
	 */
	public String getName() {
		return name;
	}

	/**
	 * Returns the value of the field.
	 * 
	 * @return Value of the field.
	 */
	public abstract T getValue();

	/**
	 * Returns the internal value of the field used for {@link #hashCode()},
	 * {@link #equals(Object)} and {@link #toString()} operations. Defaults to
	 * the value of the field.
	 * 
	 * @return Returns the value of the field.
	 */
	protected Object getInternalValue() {
		return getValue();
	}

	/**
	 * Returns the result value of the field. By default, returns the value of
	 * the field.
	 * 
	 * @return Result value of the field.
	 */
	public Object getResult() {
		return getValue();
	}

	/**
	 * Return the type of the field.
	 * 
	 * @return Type of the field.
	 */
	public abstract Class<?> getType();

	/**
	 * Accepts the annotation visitor.
	 * 
	 * @param
	 * <P>
	 * Return type of the visitor.
	 * @param visitor
	 *            visitor to accept.
	 * @return Result of the visit.
	 */
	public abstract <P> P accept(XAnnotationVisitor<P> visitor);

	/**
	 * Constructrs an annotation field with the given name.
	 * 
	 * @param name
	 *            name of the field.
	 */
	public XAnnotationField(final String name) {
		Validate.notNull(name, "Field name must not be null.");
		this.name = name;
	}

	/**
	 * Basic annotation field.
	 * 
	 * @param <T>
	 *            Type of the annotation field.
	 * @author Aleksei Valikov
	 */
	public static abstract class XField<T> extends XAnnotationField<T> {

		public XField(String name) {
			super(name);
		}

		@Override
		public String toString() {
			return getName() + "=" + getInternalValue();
		}

		@Override
		public int hashCode() {
			int hash = 0;
			final String name = getName();
			hash = hash * 37 + name.hashCode();
			final Object value = getInternalValue();
			hash = hash * 37 + value.hashCode();
			return hash;
		}

		@Override
		public boolean equals(Object obj) {
			if (!(obj instanceof XAnnotationField)) {
				return false;
			}
			if (this == obj) {
				return true;
			}
			final XAnnotationField other = (XAnnotationField) obj;
			return getName().equals(other.getName())
					&& getInternalValue().equals(other.getInternalValue());
		}
	}

	/**
	 * Annotation field with a fixed value.
	 * 
	 * @param <T>
	 *            Type of the field.
	 * @author Aleksei Valikov
	 */
	public static abstract class XFixedField<T> extends XField<T> {
		private T value;

		/**
		 * Constructs a field with the fixed value.
		 * 
		 * @param name
		 *            name of the field.
		 * @param value
		 *            value of the field.
		 */
		public XFixedField(String name, T value) {
			super(name);
			Validate.notNull(value, "Field value must not be null.");
			this.value = value;
		}

		/**
		 * Returns the fixed value of the field.
		 */
		public T getValue() {
			return value;
		}

	}

	public static class XBoolean extends XFixedField<Boolean> {
		public XBoolean(String name, Boolean value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Boolean.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitBooleanField(this);
		}
	}

	public static class XByte extends XFixedField<Byte> {
		public XByte(String name, Byte value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Byte.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitByteField(this);
		}
	}

	public static class XInt extends XFixedField<Integer> {
		public XInt(String name, Integer value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Integer.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitIntField(this);
		}
	}

	public static class XLong extends XFixedField<Long> {
		public XLong(String name, Long value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Long.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitLongField(this);
		}
	}

	public static class XShort extends XFixedField<Short> {
		public XShort(String name, Short value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Short.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitShortField(this);
		}
	}

	public static class XChar extends XFixedField<Character> {
		public XChar(String name, Character value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Character.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitCharField(this);
		}
	}

	public static class XClass extends XField<Class> {

		private Class theClass;

		private final String className;

		public XClass(String name, Class theClass) {
			super(name);
			Validate.notNull(theClass, "Class must not be null.");
			this.theClass = theClass;
			this.className = theClass.getName();
		}

		public XClass(String name, String className) {
			super(name);
			Validate.notNull(className, "Class name must not be null.");
			this.theClass = null;
			this.className = className;
		}

		public String getClassName() {
			return className;
		}

		@Override
		public Class<?> getType() {
			return Class.class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitClassField(this);
		}

		@Override
		protected Object getInternalValue() {
			return getClassName();
		}

		@Override
		public Class getValue() {
			if (this.theClass != null) {
				return this.theClass;
			} else {
				try {
					this.theClass = Class.forName(className);
					return this.theClass;
				} catch (ClassNotFoundException cnfex) {
					throw new IllegalArgumentException(cnfex);
				}
			}
		}

	}

	public static class XDouble extends XFixedField<Double> {
		public XDouble(String name, Double value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Double.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitDoubleField(this);
		}
	}

	public static class XFloat extends XFixedField<Float> {
		public XFloat(String name, Float value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Float.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitFloatField(this);
		}
	}

	public static class XEnum extends XFixedField<Enum> {

		private final Class<? extends Enum> enumClass;

		public XEnum(String name, Enum value, Class<? extends Enum> enumClass) {
			super(name, value);
			Validate.notNull(enumClass, "Enum class must not be null.");
			this.enumClass = enumClass;
		}

		@Override
		public Class<?> getType() {
			return enumClass;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitEnumField(this);
		}
	}

	public static class XString extends XFixedField<String> {
		public XString(String name, String value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return String.class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitStringField(this);
		}
	}

	public static class XAnnotation extends
			XFixedField<org.jvnet.annox.model.XAnnotation> {

		private final Class<? extends Annotation> annotationClass;

		public XAnnotation(String name,
				org.jvnet.annox.model.XAnnotation value,
				Class<? extends Annotation> annotationClass) {
			super(name, value);
			Validate.notNull(annotationClass,
					"Annotation class must not be null.");
			this.annotationClass = annotationClass;
		}

		@Override
		public Class<?> getType() {
			return annotationClass;
		}

		public Class<? extends Annotation> getAnnotationClass() {
			return annotationClass;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitAnnotationField(this);
		}

		@Override
		public Object getResult() {
			return getValue().getResult();
		}
	}

	/**
	 * Defines an array field.
	 * 
	 * @param <T>
	 *            Type of the array element.
	 * @author Aleksei Valikov
	 */
	public static abstract class XArrayField<T> extends XAnnotationField<T[]> {
		public XArrayField(String name) {
			super(name);
		}

		@Override
		public String toString() {
			final StringBuffer sb = new StringBuffer();
			sb.append(getName());
			sb.append("=[");
			final Object[] values = getInternalValue();
			for (int index = 0; index < values.length; index++) {
				final Object value = values[index];
				if (index > 0) {
					sb.append(", ");
				}
				sb.append(value);
			}
			sb.append("]");
			return sb.toString();
		}

		@Override
		public int hashCode() {
			int hash = 0;
			final String name = getName();
			hash = hash * 37 + name.hashCode();
			final Object[] values = getInternalValue();
			for (int index = 0; index < values.length; index++) {
				final Object v = values[index];
				hash = hash * 37 + v.hashCode();
			}
			return hash;
		}

		@Override
		public boolean equals(Object obj) {
			if (!(obj instanceof XArrayField<?>)) {
				return false;
			}
			if (this == obj) {
				return true;
			}
			final XArrayField<?> other = (XArrayField<?>) obj;
			if (!getName().equals(other.getName()))
				return false;

			final Object[] lhs = getInternalValue();
			final Object[] rhs = other.getInternalValue();
			if (lhs == rhs) {
				return true;
			}
			if (lhs == null || rhs == null) {
				return false;
			}
			if (lhs.length != rhs.length) {
				return false;
			}
			for (int i = 0; i < lhs.length; ++i) {
				if (!(lhs[i].equals(rhs[i])))
					return false;
			}
			return true;
		}

		@Override
		public Object[] getInternalValue() {
			return getValue();
		}

		@Override
		public Object getResult() {
			return ArrayUtils.asPrimitiveArray(getValue());
		}

	}

	/**
	 * Array field with fixed value.
	 * 
	 * @param <T>
	 *            Type of the array element.
	 * @author Aleksei Valikov
	 */
	public static abstract class XFixedArrayField<T> extends XArrayField<T> {
		private T[] value;

		public XFixedArrayField(String name, T[] value) {
			super(name);
			Validate.noNullElements(value,
					"Array must not contain null values.");
			this.value = value;
		}

		public T[] getValue() {
			return value;
		}

	}

	public static class XBooleanArray extends XFixedArrayField<Boolean> {
		public XBooleanArray(String name, Boolean[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return boolean[].class;
		}

		@Override
		public <P> P accept(XAnnotationVisitor<P> visitor) {
			return visitor.visitBooleanArrayField(this);
		}
	}

	public static class XByteArray extends XFixedArrayField<Byte> {
		public XByteArray(String name, Byte[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return byte[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitByteArrayField(this);
		}
	}

	public static class XIntArray extends XFixedArrayField<Integer> {
		public XIntArray(String name, Integer[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return int[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitIntArrayField(this);
		}
	}

	public static class XLongArray extends XFixedArrayField<Long> {
		public XLongArray(String name, Long[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return long[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitLongArrayField(this);
		}
	}

	public static class XShortArray extends XFixedArrayField<Short> {
		public XShortArray(String name, Short[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return short[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitShortArrayField(this);
		}
	}

	public static class XCharArray extends XFixedArrayField<Character> {
		public XCharArray(String name, Character[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return char[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitCharArrayField(this);
		}
	}

	public static class XClassArray extends XArrayField<Class> {

		private Class[] classes;

		private final String[] classNames;

		public XClassArray(String name, Class[] classes) {
			super(name);
			Validate.noNullElements(classes);
			this.classes = classes;
			this.classNames = new String[classes.length];
			for (int index = 0; index < classes.length; index++) {
				classNames[index] = classes[index].getName();
			}
		}

		public XClassArray(String name, String[] classNames) {
			super(name);
			Validate.noNullElements(classNames);
			this.classes = null;
			this.classNames = classNames;
		}

		// @Override
		// protected Class[] getValueInternal() {
		// }

		@Override
		public Class[] getValue() {
			if (classes != null) {
				return classes;
			} else {
				try {
					final Class[] classes = new Class[classNames.length];
					for (int index = 0; index < classNames.length; index++) {
						classes[index] = Class.forName(classNames[index]);
					}
					this.classes = classes;
					return this.classes;
				} catch (ClassNotFoundException ex) {
					throw new IllegalArgumentException(ex);
				}
			}
		}

		@Override
		public Object[] getInternalValue() {
			return getClassNames();
		}

		public String[] getClassNames() {
			return classNames;
		}

		@Override
		public Class<?> getType() {
			return Class[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitClassArrayField(this);
		}
	}

	public static class XDoubleArray extends XFixedArrayField<Double> {
		public XDoubleArray(String name, Double[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return double[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitDoubleArrayField(this);
		}

	}

	public static class XFloatArray extends XFixedArrayField<Float> {
		public XFloatArray(String name, Float[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return float[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitFloatArrayField(this);
		}
	}

	public static class XEnumArray extends XFixedArrayField<Enum> {

		private final Class<? extends Enum> enumClass;

		public XEnumArray(String name, Enum[] value,
				Class<? extends Enum> enumClass) {
			super(name, value);
			Validate.notNull(enumClass, "Enum class must not be null.");
			this.enumClass = enumClass;
		}

		@Override
		public Class<?> getType() {
			return Array.newInstance(enumClass, 0).getClass();
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitEnumArrayField(this);
		}
	}

	public static class XStringArray extends XFixedArrayField<String> {
		public XStringArray(String name, String[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return String[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitStringArrayField(this);
		}
	}

	/**
	 * Annotation array field.
	 * 
	 * @author Aleksei Valikov
	 */
	public static class XAnnotationArray extends
			XFixedArrayField<org.jvnet.annox.model.XAnnotation> {

		private final Class<? extends Annotation> annotationClass;

		public XAnnotationArray(String name,
				org.jvnet.annox.model.XAnnotation[] value,
				Class<? extends Annotation> annotationClass) {
			super(name, value);
			Validate.notNull(annotationClass,
					"Annotation class must not be null.");
			this.annotationClass = annotationClass;
		}

		@Override
		public Class<?> getType() {
			return Array.newInstance(annotationClass, 0).getClass();
		}

		public Class<? extends Annotation> getAnnotationClass() {
			return annotationClass;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitAnnotationArrayField(this);
		}

		@Override
		public Object getResult() {
			final org.jvnet.annox.model.XAnnotation[] value = getValue();
			final Annotation[] result = (Annotation[]) Array.newInstance(
					annotationClass, value.length);
			for (int index = 0; index < value.length; index++) {
				final org.jvnet.annox.model.XAnnotation annotation = value[index];
				result[index] = (Annotation) annotation.getResult();
			}
			return result;
		}
	}
}
