package org.jvnet.annox.parser;

import java.lang.annotation.Annotation;
import java.lang.reflect.Array;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.apache.commons.lang.Validate;
import org.jvnet.annox.annotation.NoSuchAnnotationFieldException;
import org.jvnet.annox.model.XAnnotation;
import org.jvnet.annox.model.XAnnotationField;
import org.jvnet.annox.model.XAnnotationField.XBoolean;
import org.jvnet.annox.model.XAnnotationField.XBooleanArray;
import org.jvnet.annox.model.XAnnotationField.XByte;
import org.jvnet.annox.model.XAnnotationField.XByteArray;
import org.jvnet.annox.model.XAnnotationField.XChar;
import org.jvnet.annox.model.XAnnotationField.XCharArray;
import org.jvnet.annox.model.XAnnotationField.XClass;
import org.jvnet.annox.model.XAnnotationField.XClassArray;
import org.jvnet.annox.model.XAnnotationField.XDouble;
import org.jvnet.annox.model.XAnnotationField.XDoubleArray;
import org.jvnet.annox.model.XAnnotationField.XEnum;
import org.jvnet.annox.model.XAnnotationField.XEnumArray;
import org.jvnet.annox.model.XAnnotationField.XFloat;
import org.jvnet.annox.model.XAnnotationField.XFloatArray;
import org.jvnet.annox.model.XAnnotationField.XInt;
import org.jvnet.annox.model.XAnnotationField.XIntArray;
import org.jvnet.annox.model.XAnnotationField.XLong;
import org.jvnet.annox.model.XAnnotationField.XLongArray;
import org.jvnet.annox.model.XAnnotationField.XShort;
import org.jvnet.annox.model.XAnnotationField.XShortArray;
import org.jvnet.annox.model.XAnnotationField.XString;
import org.jvnet.annox.model.XAnnotationField.XStringArray;
import org.jvnet.annox.util.ClassUtils;
import org.jvnet.annox.util.AnnotationElementUtils;
import org.jvnet.annox.util.ObjectUtils;
import org.jvnet.annox.util.ValueParseException;
import org.w3c.dom.Element;

public abstract class XAnnotationFieldParser<T, V> {

	public abstract XAnnotationField<T> parse(Element element, String name,
			Class<?> type) throws AnnotationElementParseException;

	public abstract XAnnotationField<T> parse(Annotation annotation,
			String name, Class<?> type) throws NoSuchAnnotationFieldException;

	public abstract XAnnotationField<T> construct(String name, V value,
			Class<?> type);

	@SuppressWarnings("unchecked")
	public <U> U getAnnotationFieldValue(Annotation annotation, String name)
			throws NoSuchAnnotationFieldException {
		final Class<? extends Annotation> annotationClass = annotation
				.annotationType();
		try {
			final Method method = annotationClass.getMethod(name);
			final U value = (U) method.invoke(annotation);
			return value;
		} catch (NoSuchMethodException nsmex) {
			throw new NoSuchAnnotationFieldException(annotationClass, name,
					nsmex);
		} catch (IllegalAccessException iaex) {
			throw new AssertionError(iaex);
		} catch (InvocationTargetException itex) {
			throw new AssertionError(itex);
		}

	}

	public abstract static class XFieldParser<T, V> extends
			XAnnotationFieldParser<T, V> {

		@Override
		public XAnnotationField<T> parse(Element element, String name,
				Class<?> type) throws AnnotationElementParseException {
			Validate.notNull(element, "Element must not be null.");
			Validate.notNull(name, "Field name must not be null.");
			Validate.notNull(type, "Type must not be null.");
			final String draft = AnnotationElementUtils.getFieldValue(element, name);

			if (draft == null) {
				return null;
			} else {
				try {
					final V value = parse(draft, type);
					return construct(name, value, type);
				} catch (ValueParseException vpex) {
					throw new AnnotationElementParseException(element, vpex);
				}
			}
		}

		@SuppressWarnings("unchecked")
		public XAnnotationField<T> parse(Annotation annotation, String name,
				Class<?> type) throws NoSuchAnnotationFieldException {
			final V value = this.<V>getAnnotationFieldValue(annotation, name);
			return construct(name, value, type);
		}

		@SuppressWarnings("unchecked")
		public V parse(String draft, Class<?> type) throws ValueParseException {
			try {
				return (V) ObjectUtils.valueOf(type, draft);
			} catch (ClassNotFoundException cnfex) {
				throw new ValueParseException(draft, type, cnfex);
			} catch (IllegalArgumentException iaex) {
				throw new ValueParseException(draft, type, iaex);
			}

		}

		public abstract XAnnotationField<T> construct(String name, V value,
				Class<?> type);

	}

	public abstract static class XArrayFieldParser<T, V> extends
			XAnnotationFieldParser<T[], V[]> {

		@Override
		public XAnnotationField<T[]> parse(Element element, String name,
				Class<?> type) throws AnnotationElementParseException {
			final String[] draft = AnnotationElementUtils.getFieldValues(element, name);
			try {
				final V[] value = parse(draft, type.getComponentType());
				return construct(name, value, type);
			} catch (ValueParseException vpex) {
				throw new AnnotationElementParseException(element, vpex);
			}
		}

		@SuppressWarnings("unchecked")
		public XAnnotationField<T[]> parse(Annotation annotation, String name,
				Class<?> type) throws NoSuchAnnotationFieldException {

			final Object values = getAnnotationFieldValue(annotation, name);
			final int length = Array.getLength(values);

			final V[] value = (V[]) Array.newInstance(ClassUtils
					.primitiveToWrapper(type.getComponentType()), length);
			for (int index = 0; index < length; index++) {
				Object object = Array.get(values, index);
				value[index] = (V) object;
			}
			return construct(name, value, type);
		}

		public abstract XAnnotationField<T[]> construct(String name, V[] value,
				Class<?> type);

		@SuppressWarnings("unchecked")
		public V[] parse(String[] draft, Class<?> type)
				throws ValueParseException {

			try {
				return (V[]) ObjectUtils.valueOf(type, draft);
			} catch (ClassNotFoundException cnfex) {
				throw new ValueParseException(draft, type, cnfex);
			} catch (IllegalArgumentException iaex) {
				throw new ValueParseException(draft, type, iaex);
			}
		}
	}

	public static class XBooleanFieldParser extends
			XFieldParser<Boolean, Boolean> {

		@Override
		public XAnnotationField<Boolean> construct(String name, Boolean value,
				Class<?> type) {
			return new XBoolean(name, value);
		}
	}

	public static class XByteFieldParser extends XFieldParser<Byte, Byte> {

		@Override
		public XAnnotationField<Byte> construct(String name, Byte value,
				Class<?> type) {
			return new XByte(name, value);
		}
	}

	public static class XIntFieldParser extends XFieldParser<Integer, Integer> {

		@Override
		public XAnnotationField<Integer> construct(String name, Integer value,
				Class<?> type) {
			return new XInt(name, value);
		}
	}

	public static class XLongFieldParser extends XFieldParser<Long, Long> {

		@Override
		public XAnnotationField<Long> construct(String name, Long value,
				Class<?> type) {
			return new XLong(name, value);
		}
	}

	public static class XShortFieldParser extends XFieldParser<Short, Short> {

		@Override
		public XAnnotationField<Short> construct(String name, Short value,
				Class<?> type) {
			return new XShort(name, value);
		}
	}

	public static class XCharFieldParser extends
			XFieldParser<Character, Character> {

		@Override
		public XAnnotationField<Character> construct(String name,
				Character value, Class<?> type) {
			return new XChar(name, value);
		}
	}

	public static class XClassFieldParser extends XFieldParser<Class, Object> {

		@Override
		public XAnnotationField<Class> construct(String name, Object value,
				Class<?> type) {
			if (value instanceof Class<?>) {
				return new XClass(name, (Class<?>) value);
			} else if (value instanceof String) {
				return new XClass(name, (String) value);
			} else {
				throw new IllegalArgumentException("Unexpected value [" + value
						+ "].");
			}
		}

		@Override
		public Object parse(String draft, Class<?> type) {
			try {
				return super.parse(draft, type);
			} catch (ValueParseException vpex) {
				return draft;
			}
		}
	}

	public static class XDoubleFieldParser extends XFieldParser<Double, Double> {

		@Override
		public XAnnotationField<Double> construct(String name, Double value,
				Class<?> type) {
			return new XDouble(name, value);
		}
	}

	public static class XFloatFieldParser extends XFieldParser<Float, Float> {

		@Override
		public XAnnotationField<Float> construct(String name, Float value,
				Class<?> type) {
			return new XFloat(name, value);
		}
	}

	public static class XStringFieldParser extends XFieldParser<String, String> {

		@Override
		public XAnnotationField<String> construct(String name, String value,
				Class<?> type) {
			return new XString(name, value);
		}
	}

	public static class XEnumFieldParser extends XFieldParser<Enum, Enum> {

		@Override
		@SuppressWarnings("unchecked")
		public XAnnotationField<Enum> construct(String name, Enum value,
				Class<?> type) {
			return new XEnum(name, value, (Class<? extends Enum>) type);
		}
	}

	public static class XAnnotationItemFieldParser extends
			XAnnotationFieldParser<XAnnotation, Annotation> {

		@Override
		@SuppressWarnings("unchecked")
		public XAnnotationField<XAnnotation> parse(Element annotationElement,
				String name, Class<?> type)
				throws AnnotationElementParseException {

			final Element element = AnnotationElementUtils.getFieldElement(annotationElement,
					name);

			if (element == null) {
				return null;
			} else {
				try {
					final XAnnotation xannotation = XAnnotationParser.GENERIC
							.parse(element);
					return new XAnnotationField.XAnnotation(name, xannotation,
							(Class<? extends Annotation>) type);
				} catch (AnnotationElementParseException aepex) {
					throw new AnnotationElementParseException(
							annotationElement, aepex);
				}
			}
		}

		@Override
		public XAnnotationField<XAnnotation> parse(Annotation annotation,
				String name, Class<?> type)
				throws NoSuchAnnotationFieldException {
			final Annotation value = getAnnotationFieldValue(annotation, name);

			final XAnnotation xannotation = XAnnotationParser.GENERIC
					.parse(value);
			@SuppressWarnings("unchecked")
			final Class<? extends Annotation> annotationType = (Class<? extends Annotation>) type;
			return new XAnnotationField.XAnnotation(name, xannotation,
					annotationType);
		}

		@SuppressWarnings("unchecked")
		@Override
		public XAnnotationField<XAnnotation> construct(String name,
				Annotation annotation, Class<?> type) {
			final XAnnotation xannotation = XAnnotationParser.GENERIC
					.parse(annotation);
			return new XAnnotationField.XAnnotation(name, xannotation,
					(Class<? extends Annotation>) type);
		}
	}

	public static class XAnnotationXItemFieldParser extends
			XAnnotationFieldParser<XAnnotation, XAnnotation> {

		@Override
		@SuppressWarnings("unchecked")
		public XAnnotationField<XAnnotation> parse(Element annotationElement,
				String name, Class<?> type)
				throws AnnotationElementParseException {

			final Element element = AnnotationElementUtils.getFieldElement(annotationElement,
					name);

			if (element == null) {
				return null;
			} else {
				try {
					final XAnnotation xannotation = XAnnotationParser.GENERIC
							.parse(element);
					return new XAnnotationField.XAnnotation(name, xannotation,
							(Class<? extends Annotation>) type);
				} catch (AnnotationElementParseException aepex) {
					throw new AnnotationElementParseException(
							annotationElement, aepex);
				}
			}
		}

		@Override
		public XAnnotationField<XAnnotation> parse(Annotation annotation,
				String name, Class<?> type)
				throws NoSuchAnnotationFieldException {
			final Annotation value = getAnnotationFieldValue(annotation, name);
			final XAnnotation xannotation = XAnnotationParser.GENERIC
					.parse(value);
			@SuppressWarnings("unchecked")
			final Class<? extends Annotation> annotationType = (Class<? extends Annotation>) type;
			return new XAnnotationField.XAnnotation(name, xannotation,
					annotationType);
		}

		@SuppressWarnings("unchecked")
		@Override
		public XAnnotationField<XAnnotation> construct(String name,
				XAnnotation annotation, Class<?> type) {
			return new XAnnotationField.XAnnotation(name, annotation,
					annotation.getAnnotationClass());
		}
	}

	public static final XBooleanFieldParser BOOLEAN = new XBooleanFieldParser();

	public static final XByteFieldParser BYTE = new XByteFieldParser();

	public static final XIntFieldParser INT = new XIntFieldParser();

	public static final XLongFieldParser LONG = new XLongFieldParser();

	public static final XShortFieldParser SHORT = new XShortFieldParser();

	public static final XCharFieldParser CHAR = new XCharFieldParser();

	public static final XClassFieldParser CLASS = new XClassFieldParser();

	public static final XDoubleFieldParser DOUBLE = new XDoubleFieldParser();

	public static final XFloatFieldParser FLOAT = new XFloatFieldParser();

	public static final XEnumFieldParser ENUM = new XEnumFieldParser();

	public static final XStringFieldParser STRING = new XStringFieldParser();

	public static final XAnnotationFieldParser ANNOTATION = new XAnnotationItemFieldParser();

	public static final XAnnotationFieldParser XANNOTATION = new XAnnotationXItemFieldParser();

	public static class XBooleanArrayFieldParser extends
			XArrayFieldParser<Boolean, Boolean> {

		@Override
		public XAnnotationField<Boolean[]> construct(String name,
				Boolean[] value, Class<?> type) {
			return new XBooleanArray(name, value);
		}
	}

	public static class XByteArrayFieldParser extends
			XArrayFieldParser<Byte, Byte> {

		@Override
		public XAnnotationField<Byte[]> construct(String name, Byte[] value,
				Class<?> type) {
			return new XByteArray(name, value);
		}
	}

	public static class XIntArrayFieldParser extends
			XArrayFieldParser<Integer, Integer> {

		@Override
		public XAnnotationField<Integer[]> construct(String name,
				Integer[] value, Class<?> type) {
			return new XIntArray(name, value);
		}
	}

	public static class XLongArrayFieldParser extends
			XArrayFieldParser<Long, Long> {

		@Override
		public XAnnotationField<Long[]> construct(String name, Long[] value,
				Class<?> type) {
			return new XLongArray(name, value);
		}
	}

	public static class XShortArrayFieldParser extends
			XArrayFieldParser<Short, Short> {

		@Override
		public XAnnotationField<Short[]> construct(String name, Short[] value,
				Class<?> type) {
			return new XShortArray(name, value);
		}
	}

	public static class XCharArrayFieldParser extends
			XArrayFieldParser<Character, Character> {

		@Override
		public XAnnotationField<Character[]> construct(String name,
				Character[] value, Class<?> type) {
			return new XCharArray(name, value);
		}
	}

	public static class XClassArrayFieldParser extends
			XArrayFieldParser<Class, Object> {

		@Override
		public Object[] parse(String[] draft, Class<?> type)
				throws ValueParseException {
			try {
				return super.parse(draft, type);
			} catch (ValueParseException ex) {
				return draft;
			}
		}

		@Override
		public XAnnotationField<Class[]> construct(String name, Object[] value,
				Class<?> type) {
			if (value instanceof Class[]) {
				return new XClassArray(name, (Class[]) value);
			} else if (value instanceof String[]) {
				return new XClassArray(name, (String[]) value);
			} else {
				throw new IllegalArgumentException("Unexpected value [" + value
						+ "].");
			}
		}
	}

	public static class XDoubleArrayFieldParser extends
			XArrayFieldParser<Double, Double> {

		@Override
		public XAnnotationField<Double[]> construct(String name,
				Double[] value, Class<?> type) {
			return new XDoubleArray(name, value);
		}
	}

	public static class XFloatArrayFieldParser extends
			XArrayFieldParser<Float, Float> {

		@Override
		public XAnnotationField<Float[]> construct(String name, Float[] value,
				Class<?> type) {
			return new XFloatArray(name, value);
		}
	}

	public static class XStringArrayFieldParser extends
			XArrayFieldParser<String, String> {

		@Override
		public XAnnotationField<String[]> construct(String name,
				String[] value, Class<?> type) {
			return new XStringArray(name, value);
		}
	}

	public static class XEnumArrayFieldParser extends
			XArrayFieldParser<Enum, Enum> {

		@SuppressWarnings("unchecked")
		@Override
		public XAnnotationField<Enum[]> construct(String name, Enum[] value,
				Class<?> type) {
			return new XEnumArray(name, value, (Class<? extends Enum>) type
					.getComponentType());
		}

	}

	public static class XAnnotationItemArrayFieldParser extends
			XAnnotationFieldParser<XAnnotation[], Annotation[]> {

		@Override
		public XAnnotationField<XAnnotation[]> parse(Element annotationElement,
				String name, Class<?> type)
				throws AnnotationElementParseException {

			final Element[] elements = AnnotationElementUtils.getFieldElements(
					annotationElement, name);

			final XAnnotation[] annotations = new XAnnotation[elements.length];

			for (int index = 0; index < elements.length; index++) {
				try {
					annotations[index] = XAnnotationParser.GENERIC
							.parse(elements[index]);
				} catch (AnnotationElementParseException aepex) {
					throw new AnnotationElementParseException(
							annotationElement, aepex);
				}
			}
			return new XAnnotationField.XAnnotationArray(name, annotations,
					(Class<? extends Annotation>) type.getComponentType());
		}

		@SuppressWarnings("unchecked")
		public XAnnotationField<XAnnotation[]> parse(Annotation annotation,
				String name, Class<?> type)
				throws NoSuchAnnotationFieldException {
			final Annotation[] value = getAnnotationFieldValue(annotation, name);
			final XAnnotation[] annotations = new XAnnotation[value.length];

			for (int index = 0; index < value.length; index++) {
				annotations[index] = XAnnotationParser.GENERIC
						.parse(value[index]);
			}
			return new XAnnotationField.XAnnotationArray(name, annotations,
					(Class<? extends Annotation>) type.getComponentType());
		}

		@SuppressWarnings("unchecked")
		public XAnnotationField<XAnnotation[]> construct(String name,
				final Annotation[] annotations, Class<?> type) {
			final XAnnotation[] xannotations = new XAnnotation[annotations.length];

			for (int index = 0; index < xannotations.length; index++) {
				xannotations[index] = XAnnotationParser.GENERIC
						.parse(annotations[index]);
			}
			return new XAnnotationField.XAnnotationArray(name, xannotations,
					(Class<? extends Annotation>) type.getComponentType());
		}
	}

	public static class XAnnotationItemXArrayFieldParser extends
			XAnnotationFieldParser<XAnnotation[], XAnnotation[]> {

		@Override
		public XAnnotationField<XAnnotation[]> parse(Element annotationElement,
				String name, Class<?> type)
				throws AnnotationElementParseException {

			final Element[] elements = AnnotationElementUtils.getFieldElements(
					annotationElement, name);

			final XAnnotation[] annotations = new XAnnotation[elements.length];

			for (int index = 0; index < elements.length; index++) {
				try {

					annotations[index] = XAnnotationParser.GENERIC
							.parse(elements[index]);
				} catch (AnnotationElementParseException aepex) {
					throw new AnnotationElementParseException(
							annotationElement, aepex);
				}
			}
			return new XAnnotationField.XAnnotationArray(name, annotations,
					(Class<? extends Annotation>) type.getComponentType());
		}

		@SuppressWarnings("unchecked")
		public XAnnotationField<XAnnotation[]> parse(Annotation annotation,
				String name, Class<?> type)
				throws NoSuchAnnotationFieldException {
			final Annotation[] value = getAnnotationFieldValue(annotation, name);

			final XAnnotation[] annotations = new XAnnotation[value.length];

			for (int index = 0; index < value.length; index++) {
				annotations[index] = XAnnotationParser.GENERIC
						.parse(value[index]);
			}
			return new XAnnotationField.XAnnotationArray(name, annotations,
					(Class<? extends Annotation>) type.getComponentType());
		}

		@SuppressWarnings("unchecked")
		public XAnnotationField<XAnnotation[]> construct(String name,
				final XAnnotation[] annotations, Class<?> type) {
			if (annotations.length == 0) {
				throw new IllegalArgumentException(
						"Could not create an annotation array field from "
								+ "the empty xannotation array.");
			}

			final Class<? extends Annotation> annotationClass = annotations[0]
					.getAnnotationClass();
			{
				for (XAnnotation annotation : annotations) {
					if (!annotationClass
							.equals(annotation.getAnnotationClass())) {
						throw new IllegalArgumentException(
								"All of the xannotations in the xannotation array "
										+ "must have the same annotation classe.");
					}
				}
			}

			final XAnnotation[] xannotations = new XAnnotation[annotations.length];

			for (int index = 0; index < xannotations.length; index++) {
				xannotations[index] = annotations[index];
			}
			return new XAnnotationField.XAnnotationArray(name, xannotations,
					annotationClass);
		}
	}

	public static final XBooleanArrayFieldParser BOOLEAN_ARRAY = new XBooleanArrayFieldParser();

	public static final XByteArrayFieldParser BYTE_ARRAY = new XByteArrayFieldParser();

	public static final XIntArrayFieldParser INT_ARRAY = new XIntArrayFieldParser();

	public static final XLongArrayFieldParser LONG_ARRAY = new XLongArrayFieldParser();

	public static final XShortArrayFieldParser SHORT_ARRAY = new XShortArrayFieldParser();

	public static final XCharArrayFieldParser CHAR_ARRAY = new XCharArrayFieldParser();

	public static final XClassArrayFieldParser CLASS_ARRAY = new XClassArrayFieldParser();

	public static final XDoubleArrayFieldParser DOUBLE_ARRAY = new XDoubleArrayFieldParser();

	public static final XFloatArrayFieldParser FLOAT_ARRAY = new XFloatArrayFieldParser();

	public static final XEnumArrayFieldParser ENUM_ARRAY = new XEnumArrayFieldParser();

	public static final XStringArrayFieldParser STRING_ARRAY = new XStringArrayFieldParser();

	public static final XAnnotationItemArrayFieldParser ANNOTATION_ARRAY = new XAnnotationItemArrayFieldParser();

	public static final XAnnotationItemXArrayFieldParser XANNOTATION_ARRAY = new XAnnotationItemXArrayFieldParser();

	@SuppressWarnings("unchecked")
	public static XAnnotationFieldParser detectType(Class theClass) {
		if (theClass == null)
			throw new IllegalArgumentException("Class must not be null.");

		if (theClass.isArray()) {
			final XAnnotationFieldParser componentType = detectType(theClass
					.getComponentType());
			if (componentType == BOOLEAN)
				return BOOLEAN_ARRAY;
			else if (componentType == BYTE)
				return BYTE_ARRAY;
			else if (componentType == INT)
				return INT_ARRAY;
			else if (componentType == LONG)
				return LONG_ARRAY;
			else if (componentType == SHORT)
				return SHORT_ARRAY;
			else if (componentType == CHAR)
				return CHAR_ARRAY;
			else if (componentType == CLASS)
				return CLASS_ARRAY;
			else if (componentType == DOUBLE)
				return DOUBLE_ARRAY;
			else if (componentType == FLOAT)
				return FLOAT_ARRAY;
			else if (componentType == ENUM)
				return ENUM_ARRAY;
			else if (componentType == STRING)
				return STRING_ARRAY;
			else if (componentType == ANNOTATION)
				return ANNOTATION_ARRAY;
			else if (componentType == XANNOTATION)
				return XANNOTATION_ARRAY;
			else
				throw new IllegalArgumentException(
						"Unknown annotation field type.");

		} else {
			if (Boolean.class.equals(theClass) || Boolean.TYPE.equals(theClass)) {
				return BOOLEAN;
			} else if (Byte.class.equals(theClass)
					|| Byte.TYPE.equals(theClass)) {
				return BYTE;
			} else if (Integer.class.equals(theClass)
					|| Integer.TYPE.equals(theClass)) {
				return INT;
			} else if (Long.class.equals(theClass)
					|| Long.TYPE.equals(theClass)) {
				return LONG;
			} else if (Short.class.equals(theClass)
					|| Short.TYPE.equals(theClass)) {
				return SHORT;
			} else if (Character.class.equals(theClass)
					|| Character.TYPE.equals(theClass)) {
				return CHAR;
			} else if (Double.class.equals(theClass)
					|| Double.TYPE.equals(theClass)) {
				return DOUBLE;
			} else if (Float.class.equals(theClass)
					|| Float.TYPE.equals(theClass)) {
				return FLOAT;
			} else if (Class.class.equals(theClass)) {
				return CLASS;
			} else if (String.class.equals(theClass)) {
				return STRING;
			} else if (Enum.class.isAssignableFrom(theClass)) {
				return ENUM;
			} else if (Annotation.class.isAssignableFrom(theClass)) {
				return ANNOTATION;
			} else if (XAnnotation.class.isAssignableFrom(theClass)) {
				return XANNOTATION;
			} else {
				throw new IllegalArgumentException(
						"Unknown annotation field type [" + theClass.getName()
								+ "].");
			}
		}
	}

	public static class XGenericFieldParser<T, V> extends
			XAnnotationFieldParser<T, V> {

		@Override
		public XAnnotationField<T> parse(Element element, String name,
				Class<?> type) throws AnnotationElementParseException {

			final XAnnotationFieldParser<T, V> parser = detectType(type);
			return parser.parse(element, name, type);
		}

		@Override
		public XAnnotationField<T> parse(Annotation annotation, String name,
				Class<?> type) throws NoSuchAnnotationFieldException {
			final XAnnotationFieldParser<T, V> parser = detectType(type);
			return parser.parse(annotation, name, type);
		}

		@Override
		public XAnnotationField<T> construct(String name, V value, Class<?> type) {
			final XAnnotationFieldParser<T, V> parser = detectType(type);
			return parser.construct(name, value, type);
		}

	}

	public static final XGenericFieldParser GENERIC = new XGenericFieldParser();
}
