package org.jvnet.annox.parser;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;
import org.jvnet.annox.Constants;
import org.jvnet.annox.annotation.AnnotationClassNotFoundException;
import org.jvnet.annox.annotation.NoSuchAnnotationFieldException;
import org.jvnet.annox.model.XAnnotation;
import org.jvnet.annox.model.XAnnotationField;
import org.w3c.dom.Element;

public class XAnnotationParser {

	public static final XAnnotationParser GENERIC = new XAnnotationParser();

	public XAnnotation parse(final Annotation annotation) {
		Validate.notNull(annotation, "Annotation must not be null.");

		final Class<? extends Annotation> annotationClass = annotation
				.annotationType();

		final XAnnotationField<?>[] fields = parseFields(annotation);

		return new XAnnotation(annotationClass, fields);

	}

	public XAnnotation[] parse(final Annotation[] annotations) {
		Validate.noNullElements(annotations,
				"Annotations must not contain nulls.");
		final XAnnotation[] xannotations = new XAnnotation[annotations.length];
		for (int index = 0; index < annotations.length; index++) {
			final Annotation annotation = annotations[index];
			xannotations[index] = parse(annotation);
		}
		return xannotations;
	}

	@SuppressWarnings("unchecked")
	public XAnnotation parse(final Element annotationElement)
			throws AnnotationElementParseException {
		Validate.notNull(annotationElement,
				"Annotation element must not be null.");

		final String name = annotationElement.getLocalName();

		final String classAttribute = annotationElement.getAttributeNS(
				Constants.NAMESPACE_URI, "class");

		final String className;
		if (!StringUtils.isEmpty(classAttribute)) {
			className = classAttribute;
		} else {
			final String namespaceURI = annotationElement.getNamespaceURI();

			if (namespaceURI != null
					&& namespaceURI.startsWith(Constants.NAMESPACE_URI_PREFIX)) {
				final String containerPrefix = namespaceURI
						.substring(Constants.NAMESPACE_URI_PREFIX.length());
				className = containerPrefix + "." + name;
			} else {
				className = name;
			}
		}

		try {
			final Class<?> draftClass = Class.forName(className);

			if (!Annotation.class.isAssignableFrom(draftClass))
				throw new AnnotationElementParseException(annotationElement,
						new IllegalArgumentException("The class ["
								+ draftClass.getName()
								+ "] is not an annotation class."));

			final Class<? extends Annotation> annotationClass = (Class<? extends Annotation>) draftClass;

			final XAnnotationField<?>[] fields = parseFields(annotationElement,
					annotationClass);
			return new XAnnotation(annotationClass, fields);
		} catch (ClassNotFoundException cnfex) {
			throw new AnnotationElementParseException(annotationElement,
					new AnnotationClassNotFoundException(className, cnfex));
		} catch (AnnotationElementParseException cnfex) {
			throw new AnnotationElementParseException(annotationElement,
					new AnnotationClassNotFoundException(className, cnfex));
		}

	}

	public XAnnotation[] parse(final Element[] annotationElements)
			throws AnnotationElementParseException {
		Validate.noNullElements(annotationElements,
				"Annotation elements must not contain null.");
		final XAnnotation[] xannotations = new XAnnotation[annotationElements.length];
		for (int index = 0; index < annotationElements.length; index++) {
			final Element annotationElement = annotationElements[index];
			try {
				xannotations[index] = parse(annotationElement);
			} catch (AnnotationElementParseException aepex) {
				throw new AnnotationElementParseException(annotationElement,
						aepex);
			}
		}
		return xannotations;
	}

	public XAnnotationField<?>[] parseFields(Element annotationElement,
			Class<? extends Annotation> annotationClass)
			throws AnnotationElementParseException {

		final Method[] methods = annotationClass.getMethods();

		final Map<String, Class<?>> fieldsMap = new HashMap<String, Class<?>>();

		for (final Method method : methods) {
			if (!Annotation.class.equals(method.getDeclaringClass())) {
				final String name = method.getName();
				final Class<?> type = method.getReturnType();
				fieldsMap.put(name, type);
			}
		}

		final List<XAnnotationField<?>> fields = new ArrayList<XAnnotationField<?>>(
				fieldsMap.size());

		for (Entry<String, Class<?>> entry : fieldsMap.entrySet()) {
			final String name = entry.getKey();
			final Class<?> type = entry.getValue();
			try {
				final XAnnotationField<?> field = parseField(annotationElement,
						name, type);
				if (field != null) {
					fields.add(field);
				}
			} catch (AnnotationElementParseException aepex) {
				throw new AnnotationElementParseException(annotationElement,
						aepex);

			}

		}
		return fields.toArray(new XAnnotationField<?>[fields.size()]);
	}

	public XAnnotationField<?>[] parseFields(Annotation annotation) {

		final Class<? extends Annotation> annotationClass = annotation
				.annotationType();
		final Method[] methods = annotationClass.getMethods();

		final List<XAnnotationField<?>> fields = new ArrayList<XAnnotationField<?>>(
				methods.length);

		for (final Method method : methods) {
			if (!Annotation.class.equals(method.getDeclaringClass())) {
				final String name = method.getName();
				final Class type = method.getReturnType();
				try {
					fields.add(parseField(annotation, name, type));
				} catch (NoSuchAnnotationFieldException nsafex) {
					// We're sure that method exists, exception can not happen
					throw new AssertionError(nsafex);
				}
			}
		}
		return fields.toArray(new XAnnotationField<?>[fields.size()]);
	}

	@SuppressWarnings("unchecked")
	public XAnnotationField<?> parseField(Element annotationElement,
			String name, Class type) throws AnnotationElementParseException {

		final XAnnotationField field = XAnnotationFieldParser.GENERIC.parse(
				annotationElement, name, type);
		return field;
	}

	@SuppressWarnings("unchecked")
	public XAnnotationField<?> parseField(Annotation annotation, String name,
			Class type) throws NoSuchAnnotationFieldException {

		final XAnnotationField field = XAnnotationFieldParser.GENERIC.parse(
				annotation, name, type);
		return field;
	}

}
